Package: S4Arrays
Title: Foundation of array-like containers in Bioconductor
Description: The S4Arrays package defines the Array virtual class to be
	extended by other S4 classes that wish to implement a container
	with an array-like semantic. It also provides: (1) low-level
	functionality meant to help the developer of such container to
	implement basic operations like display, subsetting, or coercion
	of their array-like objects to an ordinary matrix or array, and
	(2) a framework that facilitates block processing of array-like
	objects (typically on-disk objects).
biocViews: Infrastructure, DataRepresentation
URL: https://bioconductor.org/packages/S4Arrays
BugReports: https://github.com/Bioconductor/S4Arrays/issues
Version: 1.12.0
License: Artistic-2.0
Encoding: UTF-8
Authors@R: c(
	person("Hervé", "Pagès", role=c("aut", "cre"),
		email="hpages.on.github@gmail.com",
		comment=c(ORCID="0009-0002-8272-4522")),
	person("Jacques", "Serizay", role="ctb"))
Depends: R (>= 4.3.0), methods, Matrix, abind, BiocGenerics (>=
        0.45.2), S4Vectors (>= 0.47.6), IRanges
Imports: stats
LinkingTo: S4Vectors
Suggests: BiocParallel, SparseArray (>= 0.0.4), DelayedArray,
        HDF5Array, testthat, knitr, rmarkdown, BiocStyle
VignetteBuilder: knitr
Collate: utils.R rowsum.R abind.R aperm2.R array_selection.R
        Nindex-utils.R arep.R array_recycling.R Array-class.R
        dim-tuning-utils.R Array-subsetting.R Array-subassignment.R
        ArrayGrid-class.R mapToGrid.R extract_array.R type.R
        is_sparse.R read_block.R write_block.R show-utils.R
        Array-kronecker-methods.R zzz.R
git_url: https://git.bioconductor.org/packages/S4Arrays
git_branch: RELEASE_3_23
git_last_commit: b1246fd
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.23
Date/Publication: 2026-04-28
NeedsCompilation: yes
Packaged: 2026-04-29 00:22:06 UTC; biocbuild
Author: Hervé Pagès [aut, cre] (ORCID: <https://orcid.org/0009-0002-8272-4522>),
  Jacques Serizay [ctb]
Maintainer: Hervé Pagès <hpages.on.github@gmail.com>
