# S4Arrays [<img src="https://bioconductor.org/images/logo/jpg/bioconductor_logo_rgb.jpg" width="200" align="right"/>](https://bioconductor.org/)

[![in_Bioc](https://bioconductor.org/shields/years-in-bioc/S4Arrays.svg)](https://bioconductor.org/packages/S4Arrays)
[![rank](https://bioconductor.org/shields/downloads/release/S4Arrays.svg)](https://bioconductor.org/packages/stats/bioc/S4Arrays)

**S4Arrays** is an R/Bioconductor package that provides the foundation of array-like containers in Bioconductor.

See <https://bioconductor.org/packages/S4Arrays> for more information,
including for how to install the release version of the package (please refrain
from installing directly from GitHub).

|BioC version|Status|
|------------|------|
| release | [![build](https://bioconductor.org/shields/build/release/bioc/S4Arrays.svg)](https://bioconductor.org/checkResults/release/bioc-LATEST/S4Arrays/) [![platforms](https://bioconductor.org/shields/availability/release/S4Arrays.svg)](https://bioconductor.org/packages/release/S4Arrays#archives) [![dependencies](https://bioconductor.org/shields/dependencies/release/S4Arrays.svg)](https://bioconductor.org/packages/release/S4Arrays#since) |
| devel   | [![build](https://bioconductor.org/shields/build/devel/bioc/S4Arrays.svg)](https://bioconductor.org/checkResults/devel/bioc-LATEST/S4Arrays/) [![platforms](https://bioconductor.org/shields/availability/devel/S4Arrays.svg)](https://bioconductor.org/packages/devel/S4Arrays#archives) [![dependencies](https://bioconductor.org/shields/dependencies/devel/S4Arrays.svg)](https://bioconductor.org/packages/devel/S4Arrays#since) |

