\name{dim-tuning-utils}

\alias{dim-tuning-utils}
\alias{tune_Array_dims}
\alias{drop,Array-method}
\alias{dim<-,Array-method}

\title{Internal "dim tuning" utilities}

\description{
  Internal "dim tuning" utilities not meant to be used directly by the
  end user.
}

\keyword{internal}
